#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/fs.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include "clultrainit.h"
#include <asm/fasttimer.h>
#include <linux/unistd.h>

#include "camera.h"

// Information
MODULE_AUTHOR("Ernst Mayerhofer");
MODULE_DESCRIPTION("Driver for Creative Live! Ultra Webcam via USB");
MODULE_LICENSE("GPL");

#define VENDOR_ID   0x041e
#define PRODUCT_ID  0x403c
#define BUF_LEN  80
// Prototypes

static int __init clultra_init(void);
static void __exit clultra_exit(void);
static void clultra_disconnect(struct usb_interface *interface);
static void clultra_probe(struct usb_device *dev, unsigned int ifnum, const struct usb_device_id *id);
static void print_string (char *str);
static void debug(int debugmode);
static void sprinthex(char* toString, unsigned char *str,int len);
// device file
int device_ioctl(struct inode *inode,struct file *file,unsigned int ioctl_num,unsigned long ioctl_param);


static int device_open(struct inode *, struct file *);
static ssize_t device_read(struct file *, char *, size_t, loff_t *);
static int device_release(struct inode *, struct file *);


// variables
struct usb_device *camera;
static char msg[100];
static int cl_pid = 0x403c;
static int cl_vid = 0x041e;
static int globret = 0;
static unsigned int lauf = 0;
// device file
static int Major;
static int Device_Open;
static char mstring[100];
static char *msg_Ptr;
static unsigned char *mybuffer;
static unsigned char ganzesbild[0x4b008];
struct file_operations fops = {
    .read = device_read,
    .open = device_open,
    .release = device_release,
    .ioctl = device_ioctl
};
static unsigned long int read_len = 0;


static struct usb_device_id id_table [] = {
    { USB_DEVICE (VENDOR_ID,PRODUCT_ID) },
    { },
};


static struct usb_driver clultra_driver = {
    .owner = THIS_MODULE,
    .name = "clultra",
    .probe = clultra_probe,
    .disconnect = clultra_disconnect,
    .id_table = id_table,

};
MODULE_DEVICE_TABLE(usb,id_table);
static void clultra_disconnect(struct usb_interface *interface)
{
    return;
}
static void clultra_probe(struct usb_device *dev, unsigned int ifnum, const struct usb_device_id *id)
{


    if (dev->descriptor.idVendor==cl_vid && dev->descriptor.idProduct==cl_pid)
    {        
        camera = dev;

        sprintf(msg,"Found Cam: Creative Live! Ultra\n",camera->descriptor.idVendor, camera->descriptor.idProduct);
        print_string(msg);

// URB 4: Control MSG without data: IN
// want to have 8 bytes of data


init_cam(camera, msg,mybuffer,globret);

// set configuration
sprintf(msg,"Setting configuration...");
globret=usb_set_configuration(camera, 1);
debug(1);

sprintf(msg, "%i",&ganzesbild[0x4b007]-&ganzesbild[0]);
//print_string(msg);
    }
    

}
static void sprinthex(char* toString, unsigned char *str,int len)
{

    int i = 0, tmp = 0;
    for (i;i<len;i++) {
        tmp = str[i];
        sprintf(toString,"%s,%i",toString,tmp);
    }

}
static void debug(int debugmode)
{
	if (debugmode==0)
		return;
    if (globret<0)      
        sprintf(msg,"%s failed with errno: %i\n",msg,globret);
    else              
        sprintf(msg,"%s done\n",msg);
    print_string(msg);
}
static void print_string(char *str)
{
    struct tty_struct *my_tty;

    my_tty = current->tty;
    if (my_tty != NULL) {

        my_tty->driver.write (my_tty,0,str,strlen(str));    // the tty itself
//        ((my_tty->driver).write) (my_tty, 0, "\015\012", 2);
    }
}


static int device_open(struct inode *inode, struct file *file)
{
	long i;
	int raster = 1;
	int tmp = 1;
    if (Device_Open)  return -EBUSY;
    Device_Open=1;
MOD_INC_USE_COUNT;

struct urb *bild;
bild = usb_alloc_urb(0);

bild->dev = camera;
bild->pipe = usb_rcvbulkpipe(camera,0x81);
bild->transfer_buffer = ganzesbild;
//bild->transfer_flags = 2;
bild->transfer_buffer_length = 0x4b008;

usb_submit_urb(bild);

/*
for (i=0;i<307208;i++)
{
	if (tmp == 0)
		ganzesbild[i] = 0;
	else
		ganzesbild[i] = 255;
	tmp++;
	if ((i%raster)==0)
		tmp=0;
	if ((i%10000)==0)
		raster++;
}
sprintf(msg, "%i",i%2);
print_string(msg);
*/


while (bild->status==-EINPROGRESS)
	schedule_usleep(100);
schedule_usleep(1000);
sprintf(msg,"Status: %i\n",bild->status);
print_string(msg);
sprintf(msg,"Count: %i",bild->actual_length);
print_string(msg);

read_len = 0x4b008;

    return 0;
}


/*
* Called when a process, which already opened the dev file, attempts to
 * read from it.
 */
static ssize_t device_read(struct file *filp,    /* see include/linux/fs.h   */
               char *buffer,    /* buffer to fill with data */
               size_t length,    /* length of the buffer     */
               loff_t * offset)
{
    /*
     * Number of bytes actually written to the buffer
     */
	unsigned long int bytes_read=0;	
	unsigned long int bytes_offset = 0;
	
	//ganzesbild[307197] = 'b';
	//ganzesbild[307198] = 'l';
	//ganzesbild[307199] = 'a';
    /*
     * Actually put the data into the buffer
     */
	

  while (bytes_read<read_len) {

        //
        // * The buffer is in the user data segment, not the kernel
        // * segment so "*" assignment won't work.  We have to use
        // * put_user which copies data from the kernel data segment to
        // * the user data segment.
        // *
       put_user(ganzesbild[bytes_offset+bytes_read], buffer++);
	bytes_read++;


  }


    return bytes_read;

}




static int device_release(struct inode *inode, struct file *file)
{
    Device_Open=0;        /* We're now ready for our next caller */
	MOD_DEC_USE_COUNT;
    /*
     * Decrement the usage count, or else once you opened the file, you'll
     * never get get rid of the module.
     */
 //   module_put(THIS_MODULE);

    return 0;
}




/* This function is called whenever a process tries to 
 * do an ioctl on our device file. We get two extra 
 * parameters (additional to the inode and file 
 * structures, which all device functions get): the number
 * of the ioctl called and the parameter given to the 
 * ioctl function.
 *
 * If the ioctl is write or read/write (meaning output 
 * is returned to the calling process), the ioctl call 
 * returns the output of this function.
 */
int device_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
  int i;
  char *temp;
  char ch;

  /* Switch according to the ioctl called */
  switch (ioctl_num) {

    case IOCTL_GET_MSG:
      /* Give the current message to the calling 
       * process - the parameter we got is a pointer, 
       * fill it. */
      i = device_read(file, (char *) ioctl_param, 307208, 0); 
      break;

  }

  return 0;
}


int __init clultra_init()
{
    int retval = usb_register(&clultra_driver);
    if(retval)
        print_string("Usb register failed.");
    Major = register_chrdev(MAJOR_NUM, DEVICE_NAME, &fops);
        sprintf(mstring, "'mknod /dev/%s c %d 0'.\n", DEVICE_NAME, MAJOR_NUM);
        print_string(mstring);
	
	//ganzesbild = kmalloc(0x4b008,GFP_KERNEL);

    return 0;
}

void __exit clultra_exit()
{
    int ret;
    usb_deregister(&clultra_driver);

    print_string("The module has been successfully removed.\n");
    ret = unregister_chrdev(MAJOR_NUM, DEVICE_NAME);
    if (ret < 0 )
        print_string("Problem bei unregister_chrdev()");    

	//kfree(ganzesbild);
    return;
}


module_init(clultra_init);
module_exit(clultra_exit);
